package cz.cr.sara.kytara.ciselniky;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import cz.cr.sara.kytara.ciselniky.VyberAkordu.Kategorie;
import cz.cr.sara.kytara.domain.Pisen;
import cz.cr.sara.kytara.domain.Ton;
import cz.cr.sara.kytara.service.AkordManager;
import cz.cr.sara.kytara.service.PisenManager;
import cz.cr.sara.kytara.service.logic.impl.AkordPisenMapperImpl;
import cz.cr.sara.kytara.service.pojos.Akord;
import cz.cr.sara.kytara.service.pojos.AkordPisen;
import cz.cr.sara.kytara.view.AkordView;

public class PisenVyber {
  private static final Logger log = LoggerFactory.getLogger(PisenVyber.class);

  private final List<Pisen> pisne; // vsechny pisne
  private final PisenManager pisenManager;
  private final AkordManager akordManager;

  private static volatile PisenVyber INSTANCE = null;

  /**
   * Vyber akordu.
   * @param pisenManager - manazer pro akordy.
   * @throws NullPointerException
   * @return Vrati vyber akordu.
   */
  public static PisenVyber getInstance(PisenManager pisenManager,
      AkordManager akordManager) throws NullPointerException {
    if (INSTANCE == null) {
      synchronized (PisenVyber.class) {
        if (INSTANCE == null) {
          INSTANCE = new PisenVyber(pisenManager,
              akordManager);
        }
      }
    }
    return INSTANCE;
  }

  /**
   * Vyber akordu.
   * @param pisenManager - manazer pro akordy.
   * @throws NullPointerException
   */
  private PisenVyber(PisenManager pisenManager,
      AkordManager akordManager) throws NullPointerException {
    try {
      if (pisenManager == null || akordManager == null) {
        throw new NullPointerException("Chybne vstupni parametry");
      }
      this.pisenManager = pisenManager;
      this.akordManager = akordManager;
      // Seznam pisni
      this.pisne = pisenManager.getAllPisen();
      addAkordyToPisen(this.pisne);
      if (this.pisne == null || this.pisne.size() == 0) {
        throw new NullPointerException("Neexistuje seznam pisni");
      }
    }
    catch (Exception e) {
      throw new NullPointerException("PisenVyber.PisenVyber: " + e.getMessage());
    }
  }

  /**
   * @return Vrati seznam pisni.
   */
  public List<Pisen> getPisne() {
    return pisne;
  }

  /**
   * 
   * @param id - Id pisne.
   * @return Vrati pisen podle id pisne.
   */
  public Pisen getPisenById(String id) {
    Pisen pisen = null;
    for (Pisen p : pisne) {
      if (p.getId().equals(id)) {
        pisen = p;
      }
    }
    if (pisen == null) {
      pisen = pisenManager.getPisenById(id);
    }
    return pisen;
  }

  /**
   * Vrati seznam pisni s akordy.
   * @param id - Id pisne.
   * @return Vrati seznam pisni s akordy.
   */
  private void addAkordyToPisen(List<Pisen> pisne) {
    VyberAkordu vyberAkordu = getAkordVyber();
    for (Pisen pisen : pisne) {
      Set<AkordView> akordyView = new HashSet<AkordView>();
      Set<String> akordy = pisenManager.getJmenaAkorduVPisni(pisen);
      AkordPisenMapperImpl mapper = new AkordPisenMapperImpl(getTonyNazev(vyberAkordu.getTony()),
          getTypyAkordu(vyberAkordu));
      List<AkordPisen> akordyPisen = mapper.getNazevATypAkordu(akordy);
      if (log.isDebugEnabled()) {
        log.debug("akordyMapa: {}.",
            akordyPisen);
      }
      Set<Akord> vyhledaneAkordy = akordManager.getAkordyByAkordyPisen(akordyPisen);
      if (log.isDebugEnabled()) {
        log.debug("vyhledaneAkordy: {}.",
            vyhledaneAkordy);
      }
      for (Akord akord : vyhledaneAkordy) {
        AkordView av = new AkordView(akord,
            vyberAkordu);
        av.setNazev(getNazevAkorduZPisne(av,
            mapper));
        akordyView.add(av);
      }
      pisen.setAkordy(akordyView);
    }
  }

  /**
   * Vyhleda a vrati nazev akordu v pisni pomoci mapovani.
   * @param av - informace o akordu.
   * @param mapper - mapovaci trida.
   * @return Vrati nazev akordu v pisni pomoci mapovani.
   */
  private String getNazevAkorduZPisne(AkordView av,
      AkordPisenMapperImpl mapper) {
    String nazev = "";
    Map<String, String> nazevDbVersusPisen = mapper.getNazevAkordDbVersusPisen();
    for (String nazevDb : nazevDbVersusPisen.keySet()) {
      if (av.getNazev().equals(nazevDb)) {
        nazev = nazevDbVersusPisen.get(nazevDb);
      }
    }
    return nazev;
  }

  /**
   * Vrati VyberAkordu.
   * @return Vrati VyberAkordu.
   */
  private VyberAkordu getAkordVyber() {
    return AkordVyber.getInstance(akordManager).getVyberAkordu();
  }

  /**
   * Vyhleda a vrati mnozinu nazvu tonu.
   * @param tony - tony
   * @return Vrati mnozinu nazvu tonu.
   */
  private Set<String> getTonyNazev(List<Ton> tony) {
    Set<String> result = new HashSet<String>();
    for (Ton ton : tony) {
      result.add(ton.getNazev());
    }
    return result;
  }

  /**
   * Vyhleda a vrati mnozinu typu akordu.
   * @param vyberAkordu - VyberAkordu
   * @return Vrati mnozinu typu akordu.
   */
  private Set<String> getTypyAkordu(VyberAkordu vyberAkordu) {
    List<VyberAkordu.Typ> typy = getSeznamTypuAkordu(vyberAkordu);
    Set<String> result = new HashSet<String>();
    for (VyberAkordu.Typ typ : typy) {
      result.add(typ.getTyp());
    }
    return result;
  }

  /**
   * Vyhleda a vrati seznam objektu typu akordu v databazi.
   * @param vyberAkordu - VyberAkordu.
   * @return Vrati seznam vsech typu akordu v databazi.
   */
  private List<VyberAkordu.Typ> getSeznamTypuAkordu(VyberAkordu vyberAkordu) {
    List<VyberAkordu.Typ> result = new ArrayList<VyberAkordu.Typ>();
    Collection<VyberAkordu.Kategorie> kategorie = vyberAkordu.getKategorie();
    for (Kategorie k : kategorie) {
      List<VyberAkordu.Typ> typyByKategorie = k.getTypy();
      for (VyberAkordu.Typ typ : typyByKategorie) {
        result.add(typ);
      }
    }
    return result;
  }

  /**
   * Vyber pisne.
   */
  @Override
  public String toString() {
    return (pisne != null)
        ? pisne.toString()
        : "null";
  }
}
