package cz.cr.sara.kytara.repository.impl;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.EntityManagerFactory;
import javax.persistence.EntityNotFoundException;
import javax.persistence.PersistenceUnit;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import cz.cr.sara.kytara.domain.Ton;
import cz.cr.sara.kytara.domain.TypAkordu;
import cz.cr.sara.kytara.repository.AkordDAO;
import cz.cr.sara.kytara.repository.TransactionCallback;
import cz.cr.sara.kytara.repository.logic.RepositoryLogicPolohyAkordu;
import cz.cr.sara.kytara.repository.logic.RepositoryLogicUtils;
import cz.cr.sara.kytara.service.pojos.Akord;
import cz.cr.sara.kytara.service.pojos.PolohaAkordu;

/**
 * Vyhledavani akordu v DBS.
 * @author lukas
 */
@Repository("akordDao")
@Transactional(propagation = Propagation.SUPPORTS, readOnly = true)
public class AkordDAOImpl
    implements AkordDAO {
  @Autowired
  private RepositoryLogicUtils logicUtils;

  @Autowired
  private RepositoryLogicPolohyAkordu logicPolohyAkordu;

  @PersistenceUnit(unitName = "jpaEntityFactory")
  private EntityManagerFactory emf;

  private static final Logger log = LoggerFactory.getLogger(AkordDAOImpl.class);

  /**
   * Vyhleda a vrati pozadovany akord.
   * @param typAkordu - typ akordu.
   * @param nazevAkordu - nazev akordu.
   * @return Vrati pozadovany akord.
   */
  @Override
  public Akord getAkordByTypAkorduAndNazev(String typAkordu,
      String nazevAkordu) {
    // Vyhledani ID vsech tonu podle typu akordu
    List<Ton> tony = getTonyByTypAndNazev(typAkordu,
        nazevAkordu);
    // Vyhledani poloh akordu
    List<PolohaAkordu> polohyAkordu = logicPolohyAkordu.getPolohyAkorduByTony(tony,
        nazevAkordu);

    if (log.isDebugEnabled()) {
      log.debug("Vyhledane polohy: {}.",
          polohyAkordu);
    }
    return new Akord(nazevAkordu,
        typAkordu,
        tony,
        polohyAkordu);
  }

  /**
   * @return Vrati vsechny typy akordu.
   */
  @Override
  public List<TypAkordu> getAllTypAkordu() {
    Object data = new TransactionCallback() {
      @Override
      public Object runInTransaction() {
        List<TypAkordu> typyAkordu = em.createNamedQuery("TypAkordu.allTypAkordu",
            TypAkordu.class).getResultList();
        return typyAkordu;
      }
    }.run(emf);
    @SuppressWarnings("unchecked")
    List<TypAkordu> typyAkordu = (data != null)
        ? (List<TypAkordu>) data
        : null;
    if (typyAkordu == null || typyAkordu.size() == 0) {
      throw new EntityNotFoundException("Nebyly vyhledany vsechny typy akordu");
    }
    return typyAkordu;
  }

  @Override
  public List<TypAkordu> getAllTriads() {
    Object data = new TransactionCallback() {
      @Override
      public Object runInTransaction() {
        List<TypAkordu> triady = em.createNamedQuery("TypAkordu.allTriads",
            TypAkordu.class).getResultList();
        return triady;
      }
    }.run(emf);
    @SuppressWarnings("unchecked")
    List<TypAkordu> triady = (data != null)
        ? (List<TypAkordu>) data
        : null;
    if (triady == null || triady.size() == 0) {
      log.error("Nebyly vyhledany vsechny triady: {}.",
          triady);
      throw new EntityNotFoundException("Nebyly vyhledany vsechny triady");
    }
    return triady;
  }

  @Override
  public Akord getTriadaByTypAkordu(String typTriady,
      String nazev) {
    // Vyhledani ID vsech tonu podle typu akordu
    List<Ton> tony = getTonyByTypAndNazev(typTriady,
        nazev);
    // Vyhledani poloh triady
    List<PolohaAkordu> polohyTriady = logicPolohyAkordu.getPolohyTriadyByTony(tony,
        nazev);

    return new Akord(nazev,
        typTriady,
        tony,
        polohyTriady);
  }

  @Override
  public List<TypAkordu> getTypyAkorduByIntervaly(final List<Integer> intervaly) {
    // Vyhledani tonu pro vsechny ID tonu
    Object data = new TransactionCallback() {
      @Override
      public Object runInTransaction() {
        List<TypAkordu> typyAkordu = em.createNamedQuery("TypAkordu.typyByIntervaly",
            TypAkordu.class).setParameter("intervaly",
            intervaly).setParameter("pocetIntervalu",
            Long.valueOf(intervaly.size())).getResultList();
        if (typyAkordu == null) {
          throw new EntityNotFoundException("Nebyl vyhledan typ akordu pro intervaly: " + intervaly);
        }
        return typyAkordu;
      }
    }.run(emf);
    @SuppressWarnings("unchecked")
    List<TypAkordu> typyAkordu = (data != null)
        ? (List<TypAkordu>) data
        : null;
    if (log.isDebugEnabled()) {
      log.debug("Vyhledane typy akordu: {}.",
          typyAkordu);
    }
    return typyAkordu;
  }

  /**
   * Vyhleda a vrati tony podle typ a nazvu akordu.
   * @param typ - typ akordu.
   * @param nazev - nazev akordu.
   * @return Vrati tony podle typ a nazvu akordu.
   */
  private List<Ton> getTonyByTypAndNazev(String typ,
      String nazev) {
    final List<Integer> tonIds = logicUtils.getTonIdsByTypAkordu(typ,
        nazev,
        emf);
    if (log.isDebugEnabled()) {
      log.debug("Vyhledane ID tonu pro akord: {}.",
          tonIds);
    }

    // Vyhledani tonu pro vsechny ID tonu
    Object data = new TransactionCallback() {
      @Override
      public Object runInTransaction() {
        List<Ton> tony = new ArrayList<Ton>();
        for (Integer tonId : tonIds) {
          Ton ton = em.createNamedQuery("Ton.tonById",
              Ton.class).setParameter("id",
              tonId).getSingleResult();
          if (ton == null) {
            throw new EntityNotFoundException("Nebyl vyhledan ton pro ID: " + tonId);
          }
          tony.add(ton);
        }
        return tony;
      }
    }.run(emf);
    @SuppressWarnings("unchecked")
    List<Ton> tony = (data != null)
        ? (List<Ton>) data
        : null;
    if (log.isDebugEnabled()) {
      log.debug("Vyhledane tony: {}.",
          tony);
    }
    return tony;
  }
}