package cz.cr.sara.kytara.repository.logic.impl;

import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import cz.cr.sara.kytara.domain.Poloha;
import cz.cr.sara.kytara.domain.Ton;
import cz.cr.sara.kytara.repository.PolohaDAO;
import cz.cr.sara.kytara.repository.logic.PolohaAkorduIntelligence;
import cz.cr.sara.kytara.service.pojos.PolohaAkordu;

/**
 * Vyhledani a sestaveni konkretni polohy / hmatu akordu.
 * @author lukas
 */
@Component("polohaAkorduIntelligence")
public class PolohaAkorduIntelligenceImpl
    implements PolohaAkorduIntelligence {
  @Autowired
  private PolohaDAO polohaDao;

  @Autowired
  private ZapisPrazceLogic zapisPrazceLogic;

  private static final Logger log = LoggerFactory.getLogger(PolohaAkorduIntelligenceImpl.class);

  /**
   * Vyhleda a sestavi konkretni polohy / hmaty akordu.
   * @param zakladniTon - poloha akordu zakladniho tonu.
   * @param tony - seznam tonu.
   * @return Vrati seznam poloh akordu.
   */
  @Override
  public PolohaAkordu createPolohaAkordu(Poloha zakladniTon,
      List<Ton> tony) {
    // Informace o jednotlivych strunach
    PolohaAkorduIntelligenceImpl.InfoStruna[] infoStruny = getInfoOStrunach();

    if (log.isDebugEnabled()) {
      log.debug("zakladni ton: {}, tony: {}, info o strunach: {}.",
          zakladniTon,
          tony,
          infoStruny);
    }
    // Nastaveni prazcu pro jednotlive struny
    setAllPrazecAkord(zakladniTon,
        infoStruny,
        tony);

    if (log.isDebugEnabled()) {
      log.debug("info o strunach prazce: {}.",
          (Object) infoStruny);
    }
    //Nastaveni prazcu spolecna cast
    setAllPrazecCommon(zakladniTon.getPrazec(),
        infoStruny);
    if (log.isDebugEnabled()) {
      log.debug("info o strunach po logice: {}.",
          (Object) infoStruny);
    }
    // Sestaveni konkretni polohy / hmatu akordu
    return build(infoStruny);
  }

  /**
   * Vyhleda a sestavi konkretni polohy / hmaty triady.
   * @param zakladniTon - poloha triady zakladniho tonu.
   * @param tony - seznam tonu.
   * @return Vrati seznam poloh triady.
   */
  @Override
  public PolohaAkordu createPolohaTriady(Poloha zakladniTon,
      List<Ton> tony) {
    // Informace o jednotlivych strunach
    PolohaAkorduIntelligenceImpl.InfoStruna[] infoStruny = getInfoOStrunach();

    // Nastaveni prazcu pro jednotlive struny
    setAllPrazecTriada(zakladniTon,
        infoStruny,
        tony);

    //Nastaveni prazcu spolecna cast
    setAllPrazecCommon(zakladniTon.getPrazec(),
        infoStruny);

    // Sestaveni konkretni polohy / hmatu triady
    return build(infoStruny);
  }

  /**
   * Sestavi a vrati informace o jednotlivych strunach.
   * @return
   */
  private PolohaAkorduIntelligenceImpl.InfoStruna[] getInfoOStrunach() {
    PolohaAkorduIntelligenceImpl.InfoStruna[] infoStruny = new PolohaAkorduIntelligenceImpl.InfoStruna[PolohaAkorduIntelligenceImpl.Struna.values().length];
    int i = 0;
    for (PolohaAkorduIntelligenceImpl.Struna struna : PolohaAkorduIntelligenceImpl.Struna.values()) {
      infoStruny[i++] = new PolohaAkorduIntelligenceImpl.InfoStruna(struna);
    }
    return infoStruny;
  }

  /**
   * Nastaveni prazcu pro jednotlive struny u akordu. Struny pod zakladnim tonem se nebudou pouzivat, nastavi se
   * zakladni ton a vsechny kombinace prazcu na dalsich strunach.
   * @param zakladniTon - poloha akordu zakladniho tonu.
   * @param infoStruny - informace o jednotlivych strunach.
   * @param tony - seznam tonu.
   */
  private void setAllPrazecAkord(Poloha zakladniTon,
      PolohaAkorduIntelligenceImpl.InfoStruna[] infoStruny,
      List<Ton> tony) {
    // Struna zakladniho tonu
    PolohaAkorduIntelligenceImpl.Struna strunaZakladniTon = PolohaAkorduIntelligenceImpl.Struna.fromValue(zakladniTon.getStruna());
    // Poradi prazce zakladniho tonu
    int prazecZakladniTon = zakladniTon.getPrazec();

    if (strunaZakladniTon == PolohaAkorduIntelligenceImpl.Struna.STRUNA_2) {
      for (PolohaAkorduIntelligenceImpl.Struna struna : PolohaAkorduIntelligenceImpl.Struna.values()) {
        switch (struna) {
        case STRUNA_1:
          break;
        case STRUNA_2:
          infoStruny[struna.poradi].prazecZakladniTon = prazecZakladniTon;
          break;
        default:
          infoStruny[struna.poradi].prazceStruna = getPrazceByStrunaAndTons(struna,
              tony);
          break;
        }
      }
    }
    else {
      if (strunaZakladniTon == PolohaAkorduIntelligenceImpl.Struna.STRUNA_3) {
        for (PolohaAkorduIntelligenceImpl.Struna struna : PolohaAkorduIntelligenceImpl.Struna.values()) {
          switch (struna) {
          case STRUNA_1:
          case STRUNA_2:
            break;
          case STRUNA_3:
            infoStruny[struna.poradi].prazecZakladniTon = prazecZakladniTon;
            break;
          default:
            infoStruny[struna.poradi].prazceStruna = getPrazceByStrunaAndTons(struna,
                tony);
            break;
          }
        }
      }
      else {
        for (PolohaAkorduIntelligenceImpl.Struna struna : PolohaAkorduIntelligenceImpl.Struna.values()) {
          switch (struna) {
          case STRUNA_1:
            infoStruny[struna.poradi].prazecZakladniTon = prazecZakladniTon;
            break;
          default:
            infoStruny[struna.poradi].prazceStruna = getPrazceByStrunaAndTons(struna,
                tony);
            break;
          }
        }
      }
    }
  }

  /**
   * Nastaveni prazcu pro jednotlive struny u triady. Struny pod zakladnim tonem se nebudou pouzivat, nastavi se
   * zakladni ton a vsechny kombinace prazcu na dalsich strunach.
   * @param zakladniTon - poloha triady zakladniho tonu.
   * @param infoStruny - informace o jednotlivych strunach.
   * @param tony - seznam tonu.
   */
  private void setAllPrazecTriada(Poloha zakladniTon,
      PolohaAkorduIntelligenceImpl.InfoStruna[] infoStruny,
      List<Ton> tony) {
    // Struna zakladniho tonu
    PolohaAkorduIntelligenceImpl.Struna strunaZakladniTon = PolohaAkorduIntelligenceImpl.Struna.fromValue(zakladniTon.getStruna());
    // Poradi prazce zakladniho tonu
    int prazecZakladniTon = zakladniTon.getPrazec();

    if (strunaZakladniTon == PolohaAkorduIntelligenceImpl.Struna.STRUNA_2) {
      for (PolohaAkorduIntelligenceImpl.Struna struna : PolohaAkorduIntelligenceImpl.Struna.values()) {
        switch (struna) {
        case STRUNA_1:
        case STRUNA_5:
        case STRUNA_6:
          break;
        case STRUNA_2:
          infoStruny[struna.poradi].prazecZakladniTon = prazecZakladniTon;
          break;
        default:
          infoStruny[struna.poradi].prazceStruna = getPrazceByStrunaAndTons(struna,
              tony);
          break;
        }
      }
    }
    else {
      if (strunaZakladniTon == PolohaAkorduIntelligenceImpl.Struna.STRUNA_3) {
        for (PolohaAkorduIntelligenceImpl.Struna struna : PolohaAkorduIntelligenceImpl.Struna.values()) {
          switch (struna) {
          case STRUNA_1:
          case STRUNA_2:
          case STRUNA_6:
            break;
          case STRUNA_3:
            infoStruny[struna.poradi].prazecZakladniTon = prazecZakladniTon;
            break;
          default:
            infoStruny[struna.poradi].prazceStruna = getPrazceByStrunaAndTons(struna,
                tony);
            break;
          }
        }
      }
      else {
        if (strunaZakladniTon == PolohaAkorduIntelligenceImpl.Struna.STRUNA_4) {
          for (PolohaAkorduIntelligenceImpl.Struna struna : PolohaAkorduIntelligenceImpl.Struna.values()) {
            switch (struna) {
            case STRUNA_1:
            case STRUNA_2:
            case STRUNA_3:
              break;
            case STRUNA_4:
              infoStruny[struna.poradi].prazecZakladniTon = prazecZakladniTon;
              break;
            default:
              infoStruny[struna.poradi].prazceStruna = getPrazceByStrunaAndTons(struna,
                  tony);
              break;
            }
          }
        }
        else {
          for (PolohaAkorduIntelligenceImpl.Struna struna : PolohaAkorduIntelligenceImpl.Struna.values()) {
            switch (struna) {
            case STRUNA_4:
            case STRUNA_5:
            case STRUNA_6:
              break;
            case STRUNA_1:
              infoStruny[struna.poradi].prazecZakladniTon = prazecZakladniTon;
              break;
            default:
              infoStruny[struna.poradi].prazceStruna = getPrazceByStrunaAndTons(struna,
                  tony);
              break;
            }
          }
        }
      }
    }
  }

  private void setAllPrazecCommon(int prazecZakladniTon,
      PolohaAkorduIntelligenceImpl.InfoStruna[] infoStruny) {
    // Abychom si neprepsali hodnotu zakladniho prazce
    for (PolohaAkorduIntelligenceImpl.InfoStruna infoStruna : infoStruny) {
      switch (infoStruna.struna) {
      case STRUNA_1:
        break;
      case STRUNA_2:
      case STRUNA_3:
        if (infoStruna.prazecZakladniTon != prazecZakladniTon) {
          infoStruna.prazecZakladniTon = setPrazecToStruna(infoStruna.prazceStruna,
              prazecZakladniTon,
              infoStruny[infoStruna.struna.poradi + 1].prazceStruna,
              infoStruny[infoStruna.struna.poradi - 1].prazecZakladniTon);
        }
        break;
      case STRUNA_6:
        infoStruna.prazecZakladniTon = setPrazecToStruna(infoStruna.prazceStruna,
            prazecZakladniTon,
            infoStruna.prazceStruna,
            infoStruny[infoStruna.struna.poradi - 1].prazecZakladniTon);
        break;
      default:
        infoStruna.prazecZakladniTon = setPrazecToStruna(infoStruna.prazceStruna,
            prazecZakladniTon,
            infoStruny[infoStruna.struna.poradi + 1].prazceStruna,
            infoStruny[infoStruna.struna.poradi - 1].prazecZakladniTon);
        break;
      }
    }
    if (log.isDebugEnabled()) {
      for (PolohaAkorduIntelligenceImpl.InfoStruna infoStruna : infoStruny) {
        log.debug("Informace o strune: {}.",
            infoStruna);
      }
    }
  }

  /**
   * Vyhleda a vraci polohy (prazce) vsech vyskytu tonu na jedne strune.
   * @param struna - pozadovana struna.
   * @param tony - seznam tonu.
   * @return Vraci polohy (prazce) vsech vyskytu tonu na jedne strune.
   */
  private List<Integer> getPrazceByStrunaAndTons(PolohaAkorduIntelligenceImpl.Struna struna,
      List<Ton> tony) {
    return polohaDao.getPrazceByStrunaAndTony(struna.poradi,
        tony);
  }

  /**
   * Vraci vybrany prazec z nekolika moznych na jedne strune. Ma to byt ten nejvhodnejsi.
   * @param prazceNaStrune - mozne prazce na dane strune.
   * @param prazecZakladniTon - poradi prazce zakladniho tonu.
   * @param prazceNaDalsiStrune - mozne prazce na dalsi strune. Pro posledni strunu prazce na te same strune.
   * @param predchoziStrunaPrazec - poradi prazce predchozi struna.
   * @return Vrati cislo prazce. Pokud je prazec nenalezen, vraci -1.
   */
  private int setPrazecToStruna(List<Integer> prazceNaStrune,
      int prazecZakladniTon,
      List<Integer> prazceNaDalsiStrune,
      int predchoziStrunaPrazec) {
    if (prazceNaStrune != null) {
      int prazec = zapisPrazceLogic.zapisPrazceId(prazceNaStrune,
          prazecZakladniTon,
          prazceNaDalsiStrune,
          predchoziStrunaPrazec);
      if (log.isInfoEnabled()) {
        log.info("Vybrany prazec: {}.",
            prazec);
      }
      return prazec;
    }
    else
      return ZapisPrazceLogic.NEPOUZITA_STRUNA_PRAZEC_ID;
  }

  /**
   * Sestavi konkretni polohu/hmat akordu.
   * @param infoStruny - informace o jednotlivych strunach.
   * @return Vrati seznam poloh akordu.
   */
  private PolohaAkordu build(PolohaAkorduIntelligenceImpl.InfoStruna[] infoStruny) {
    boolean maSmysl = maPolohaSmysl(infoStruny);
    if (!maSmysl) {
      return new PolohaAkordu();
    }
    List<Poloha> polohy = new ArrayList<Poloha>();
    for (PolohaAkorduIntelligenceImpl.InfoStruna infoStruna : infoStruny) {
      Poloha ton = new Poloha();
      ton.setStruna(infoStruna.struna.poradi);
      ton.setPrazec(infoStruna.prazecZakladniTon);
      polohy.add(ton);
    }
    PolohaAkordu polohaAkordu = new PolohaAkordu();
    polohaAkordu.setPoloha(polohy);
    return polohaAkordu;
  }

  /**
   * Zjistuje, zda je v akordu minimalni pocet tonu.
   * @param infoStruny
   * @return
   */
  private boolean maPolohaSmysl(PolohaAkorduIntelligenceImpl.InfoStruna[] infoStruny) {
    int pocetTonuNaPrazci = 0;
    for (PolohaAkorduIntelligenceImpl.InfoStruna infoStruna : infoStruny) {
      if (infoStruna.prazecZakladniTon != ZapisPrazceLogic.NEPOUZITA_STRUNA_PRAZEC_ID) {
        pocetTonuNaPrazci++;
      }
    }
    if (pocetTonuNaPrazci < 3) {
      return false;
    }
    return true;
  }

  /**
   * Informace o jedne strune.
   * @author lukas
   */
  private static class InfoStruna {
    private final PolohaAkorduIntelligenceImpl.Struna struna; // Struna
    private int prazecZakladniTon; // Poradi prazce zakladniho tonu
    private List<Integer> prazceStruna; // Polohy (prazce) vsechny vyskyty na jedne strune

    /**
     * Informace o jedne strune.
     * @param struna - struna.
     */
    private InfoStruna(PolohaAkorduIntelligenceImpl.Struna struna) {
      this.struna = struna;
      this.prazecZakladniTon = ZapisPrazceLogic.NEPOUZITA_STRUNA_PRAZEC_ID;
    }

    /**
     * Informace o jedne strune.
     */
    @Override
    public String toString() {
      StringBuffer info = new StringBuffer();
      info.append("[STRUNA=");
      info.append(struna);
      info.append(",PORADI-PRAZCE=");
      info.append(prazecZakladniTon);
      info.append(",PRAZCE=");
      info.append(prazceStruna);
      info.append("]");
      return info.toString();
    }
  }

  /**
   * Struny.
   * @author lukas
   */
  public static enum Struna {
    STRUNA_1(0),
    STRUNA_2(1),
    STRUNA_3(2),
    STRUNA_4(3),
    STRUNA_5(4),
    STRUNA_6(5);

    private final int poradi; // Poradi struny ( od '0' )

    /**
     * Sruna.
     * @param poradi - poradi struny ( od '0' ).
     */
    private Struna(int poradi) {
      this.poradi = poradi;
    }

    /**
     * @return Vrati poradi struny.
     */
    public int getPoradi() {
      return poradi;
    }

    /**
     * Vyhleda a vrati pozadovanou strunu.
     * @param poradi - poradi struny ( od '0' ).
     * @return Vrati pozadovanou strunu.
     * @throws NullPointerException
     */
    private static Struna fromValue(int poradi) throws NullPointerException {
      try {
        for (PolohaAkorduIntelligenceImpl.Struna struna : PolohaAkorduIntelligenceImpl.Struna.values()) {
          if (struna.poradi == poradi) {
            return struna;
          }
        }
        throw new NullPointerException("Neexistuje struna pro poradi: " + poradi);
      }
      catch (Exception e) {
        throw new NullPointerException("PolohaAkorduIntelligenceImpl.Struna.fromValue: " + e.getMessage());
      }
    }

    /**
     * Struna.
     */
    @Override
    public String toString() {
      StringBuffer info = new StringBuffer(50);
      info.append("[PORADI-STRUNY=");
      info.append(poradi);
      info.append("]");
      return info.toString();
    }
  }
}
