package cz.cr.sara.kytara.repository.logic.impl;

import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import cz.cr.sara.kytara.domain.Poloha;
import cz.cr.sara.kytara.domain.Ton;
import cz.cr.sara.kytara.repository.logic.PolohaAkorduIntelligence;
import cz.cr.sara.kytara.repository.logic.RepositoryLogicPolohyAkordu;
import cz.cr.sara.kytara.service.pojos.PolohaAkordu;

/**
 * Vyhledani poloh akordu.
 * @author lukas
 */
@Component("logicPolohyAkordu")
public class RepositoryLogicPolohyAkorduImpl
    implements RepositoryLogicPolohyAkordu {
  @Autowired
  private PolohaAkorduIntelligence polohaAkorduIntelligence;

  private final static Logger log = LoggerFactory.getLogger(RepositoryLogicPolohyAkorduImpl.class);

  /**
   * Vyhledani poloh akordu.
   * @param tony - seznam tonu.
   * @param prvniTonNazev - nazev prvniho tonu.
   * @return Vrati polohy akordu.
   */
  @Override
  public List<PolohaAkordu> getPolohyAkorduByTony(List<Ton> tony,
      String prvniTonNazev) {
    List<PolohaAkordu> polohyAkordu = new ArrayList<PolohaAkordu>();

    // Vyhledani prvniho tonu akordu
    Ton prvniTon = getPrvniTonAkordu(tony,
        prvniTonNazev);
    if (prvniTon != null) {
      if (log.isDebugEnabled()) {
        log.debug("prvniTon: {}",
            prvniTon);
      }
      // Seznam poloh pouze na prvni, druhe a treti strune
      List<Poloha> triPolohyByPrvniTon = getPolohyOnFirstThreeStrings(prvniTon);
      if (log.isDebugEnabled()) {
        log.debug("Prvni ton 3 polohy: {}",
            triPolohyByPrvniTon);
      }
      PolohaAkordu polohaAkordu = null;
      for (Poloha zakladniTon : triPolohyByPrvniTon) {
        polohaAkordu = polohaAkorduIntelligence.createPolohaAkordu(zakladniTon,
            tony);
        if (polohaAkordu.getPoloha() != null) {
          polohyAkordu.add(polohaAkordu);
        }
        if (log.isInfoEnabled()) {
          log.info("Poloha akordu: {}",
              polohaAkordu);
        }
      }
    }
    return polohyAkordu;
  }

  /**
   * Vyhleda a vrati prvni ton akordu.
   * @param tony - seznam tonu.
   * @param prvniTonNazev - nazev prvniho tonu.
   * @return Vrati prvni ton akordu.
   */
  private Ton getPrvniTonAkordu(List<Ton> tony,
      String prvniTonNazev) {
    for (Ton ton : tony) {
      if (ton.getNazev().equals(prvniTonNazev)) {
        return ton;
      }
    }
    return null;
  }

  /**
   * Vyhleda a vrati seznam poloh tonu pouze na prvni, druhe a treti strune. Polohy budou serazeny podle nejmensiho
   * prazce.
   * @param prvniTon - prvni ton.
   * @return Vrati seznam poloh pouze na prvni, druhe a treti strune.
   */
  private List<Poloha> getPolohyOnFirstThreeStrings(Ton prvniTon) {
    List<Poloha> polohy = new ArrayList<Poloha>();
    for (Poloha poloha : prvniTon.getPolohy()) {
      if (poloha.getStruna() < PolohaAkorduIntelligenceImpl.Struna.STRUNA_4.getPoradi()) {
        polohy.add(poloha);
      }
    }
    return orderPolohyByPrazecId(polohy);
  }

  /**
   * Serazeni poloh podle nejmensiho prazce.
   * @param polohy - seznam poloh.
   * @return Vrati seznam poloh serazeny podle nejmensiho prazce.
   */
  private List<Poloha> orderPolohyByPrazecId(List<Poloha> polohy) {
    List<Poloha> orderedList = new ArrayList<Poloha>();
    Poloha polohaMin = polohy.get(0);
    Poloha polohaMax = polohy.get(0);
    Poloha polohaBetween = polohy.get(0);

    // Naplni polohaMin a polohaMax
    for (Poloha poloha : polohy) {
      if (poloha.getPrazec() < polohaMin.getPrazec()) {
        polohaMin = poloha;
      }
      if (poloha.getPrazec() > polohaMax.getPrazec()) {
        polohaMax = poloha;
      }
    }
    // Naplni polohaBetween
    for (Poloha poloha : polohy) {
      if (poloha.getPrazec() > polohaMin.getPrazec() && poloha.getPrazec() < polohaMax.getPrazec()) {
        polohaBetween = poloha;
      }
    }
    orderedList.add(polohaMin);
    orderedList.add(polohaBetween);
    orderedList.add(polohaMax);
    return orderedList;
  }

  @Override
  public List<PolohaAkordu> getPolohyTriadyByTony(List<Ton> tony,
      String prvniTonNazev) {
    List<PolohaAkordu> polohyTriady = new ArrayList<PolohaAkordu>();

    // Vyhledani prvniho tonu triady
    Ton prvniTon = getPrvniTonAkordu(tony,
        prvniTonNazev);
    if (prvniTon != null) {
      if (log.isDebugEnabled()) {
        log.debug("prvniTon: {}",
            prvniTon);
      }
      // Seznam poloh pouze na prvni, druhe, treti a ctvrte strune
      List<Poloha> ctyriPolohyByPrvniTon = getPolohyOnFirstFourStrings(prvniTon);
      if (log.isDebugEnabled()) {
        log.debug("Prvni ton 3 polohy: {}",
            ctyriPolohyByPrvniTon);
      }
      PolohaAkordu polohaTriady = null;
      for (Poloha zakladniTon : ctyriPolohyByPrvniTon) {
        polohaTriady = polohaAkorduIntelligence.createPolohaTriady(zakladniTon,
            tony);
        if (polohaTriady.isPoloha() && polohaTriady.getPoloha().size() != 0) {
          polohyTriady.add(polohaTriady);
        }
        if (log.isDebugEnabled()) {
          log.debug("Poloha triady prazec na 2.strune: {}",
              polohaTriady.getPoloha().get(1).getPrazec());
        }
      }
    }
    return polohyTriady;
  }

  /**
   * Vyhleda a vrati seznam poloh tonu pouze na prvni, druhe, treti a ctvrte strune. Polohy budou serazeny podle
   * nejmensiho prazce.
   * @param prvniTon - prvni ton.
   * @return Vrati seznam poloh pouze na prvni, druhe, treti a ctvrte strune.
   */
  private List<Poloha> getPolohyOnFirstFourStrings(Ton prvniTon) {
    List<Poloha> polohy = new ArrayList<Poloha>();
    for (Poloha poloha : prvniTon.getPolohy()) {
      if (poloha.getStruna() < PolohaAkorduIntelligenceImpl.Struna.STRUNA_5.getPoradi()) {
        polohy.add(poloha);
      }
    }
    return orderPolohyByPrazecIdForTriady(polohy);
  }

  /**
   * Serazeni poloh podle nejmensiho prazce pro triady, kde se zakladni ton vybira na 4 strunach.
   * @param polohy - seznam poloh.
   * @return Vrati seznam poloh serazeny podle nejmensiho prazce.
   */
  private List<Poloha> orderPolohyByPrazecIdForTriady(List<Poloha> polohy) {
    List<Poloha> orderedList = new ArrayList<Poloha>();
    Poloha poloha1 = polohy.get(0);
    Poloha poloha2 = polohy.get(0);
    Poloha poloha3 = polohy.get(0);
    Poloha poloha4 = polohy.get(0);

    // Naplni polohaMin a polohaMax
    for (Poloha poloha : polohy) {
      if (poloha.getPrazec() < poloha1.getPrazec()) {
        poloha1 = poloha;
      }
      if (poloha.getPrazec() > poloha4.getPrazec()) {
        poloha4 = poloha;
      }
    }
    // Naplni prostredni polohy
    List<Poloha> prostredniPolohy = new ArrayList<Poloha>();
    for (Poloha poloha : polohy) {
      if (poloha.getPrazec() > poloha1.getPrazec() && poloha.getPrazec() < poloha4.getPrazec()) {
        prostredniPolohy.add(poloha);
        poloha2 = poloha;
        poloha3 = poloha;
      }
    }
    for (Poloha poloha : prostredniPolohy) {
      if (poloha.getPrazec() < poloha2.getPrazec()) {
        poloha2 = poloha;
      }
      if (poloha.getPrazec() > poloha3.getPrazec()) {
        poloha3 = poloha;
      }
    }
    orderedList.add(poloha1);
    orderedList.add(poloha2);
    orderedList.add(poloha3);
    orderedList.add(poloha4);
    return orderedList;
  }
}