package cz.cr.sara.kytara.repository.logic.impl;

import java.util.List;

import javax.persistence.EntityManagerFactory;
import javax.persistence.EntityNotFoundException;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import cz.cr.sara.kytara.repository.TransactionCallback;
import cz.cr.sara.kytara.repository.logic.RepositoryLogicUtils;

/**
 * Logicke operace pri vyhledavani akordu.
 * @author lukas
 */
@Component("logicUtils")
public class RepositoryLogicUtilsImpl
    implements RepositoryLogicUtils {
  @Autowired
  private RepositoryLogicHelper logicHelper;

  /**
   * Vyhleda a vraci ID vsech tonu podle typu akordu.
   * @param typAkordu - typ akordu.
   * @param nazevTon - nazev tonu.
   * @param emf - tovarna entit pro praci s DBS.
   * @return Vraci ID vsech tonu podle typu akordu.
   */
  @Override
  public List<Integer> getTonIdsByTypAkordu(final String typAkordu,
      final String nazevTon,
      final EntityManagerFactory emf) {
    // Vyhledani intervalu pro typ akordu
    Object data = new TransactionCallback() {
      @Override
      public Object runInTransaction() {
        List<Integer> intervaly = em.createNamedQuery("Interval.findByTypAkordu",
            Integer.class).setParameter("typ",
            typAkordu).getResultList();
        return intervaly;
      }
    }.run(emf);
    if (data == null) {
      throw new EntityNotFoundException("Nebyly vyhledany zadne intervaly pro typ akordu: " + typAkordu);
    }
    @SuppressWarnings("unchecked")
    List<Integer> intervaly = (data != null)
        ? (List<Integer>) data
        : null;

    // Vyhledani ID zakladniho tonu podle jeho nazvu
    int zakladniTonId = getTonIdByTonNazev(nazevTon,
        emf);

    // Seznam ID tonu
    List<Integer> tonIds = validateTonIds(intervaly,
        zakladniTonId);
    if (tonIds == null || tonIds.size() == 0) {
      throw new EntityNotFoundException("Nebyl vyhledan seznam ID tonu pro akord: " + nazevTon + " / " + typAkordu);
    }
    return tonIds;
  }

  /**
   * Vyhleda a vraci ID tonu podle jeho nazvu.
   * @param nazevTon - nazev tonu.
   * @param emf - tovarna entit pro praci s DBS.
   * @return Vraci ID tonu podle jeho nazvu.
   */
  private int getTonIdByTonNazev(final String nazevTon,
      EntityManagerFactory emf) {
    Object tonId = new TransactionCallback() {
      @Override
      public Object runInTransaction() {
        return em.createNamedQuery("Ton.idTonByNazev",
            Integer.class).setParameter("nazev",
            nazevTon).getSingleResult();
      }
    }.run(emf);
    if (tonId == null) {
      throw new EntityNotFoundException("Nebyl vyhledan ID tonu podle jeho nazvu: " + nazevTon);
    }
    return (int) tonId;
  }

  /**
   * Vytvori a vrati seznam ID tonu.
   * @param intervaly - seznam vzdalenosti od zakladniho tonu.
   * @param zakladniTonId - ID zakladniho tonu.
   * @return Vrati seznam ID tonu.
   */
  private List<Integer> validateTonIds(List<Integer> intervaly,
      int zakladniTonId) {
    List<Integer> validatedIntervals = logicHelper.validateIntervals(intervaly);
    return logicHelper.validateTonIds(zakladniTonId,
        validatedIntervals);
  }
}
