package cz.cr.sara.kytara.repository.logic.impl;

import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

/**
 * Vyber spravneho prazce na jedne strune.
 * @author lukas
 */
@Component("zapisPrazceLogic")
public class ZapisPrazceLogic {
  /** Maximalni odchylka od zakladniho prazce */
  public static final int DEFAULT_POSUN = 2;
  /** ID pro strunu, ktera se nebude drzet */
  private static final int PRAZDNA_STRUNA_PRAZEC_ID = 0;
  /** ID pro strunu, na ktere neni zadny prazec pouzitelny. Struna se nehraje. */
  public static final int NEPOUZITA_STRUNA_PRAZEC_ID = -1;

  private int maxPosun = ZapisPrazceLogic.DEFAULT_POSUN;

  private static final Logger log = LoggerFactory.getLogger(ZapisPrazceLogic.class);

  /**
   * Vnitrni logika pro vyber spravneho prazce na jedne strune.
   * @param prazceNaStrune - seznam prazcu na strune.
   * @param zakladniPrazec - zakladni prazec.
   * @param prazceNaDalsiStrune - seznam prazcu na dalsi strune.
   * @param predchoziStrunaPrazec - prazec predchozi struna.
   * @return Vrati ID prazce.
   */
  public int zapisPrazceId(List<Integer> prazceNaStrune,
      int zakladniPrazec,
      List<Integer> prazceNaDalsiStrune,
      int predchoziStrunaPrazec) {
    if (log.isDebugEnabled()) {
      log.debug("zakladniPrazec: {}.",
          zakladniPrazec);
      log.debug("prazceStruna: {}.",
          prazceNaStrune);
    }

    int result = 0;
    Integer[] prazce = new Integer[prazceNaStrune.size()];
    prazceNaStrune.toArray(prazce);

    switch (zakladniPrazec) {
    case 0:
      maxPosun++;
    case 1:
    case 2:
    case 3:
      result = vyberSPrazdouStrunou(prazce,
          zakladniPrazec,
          prazceNaDalsiStrune,
          predchoziStrunaPrazec);
      break;

    default:
      result = spolecnaPravidla(prazce,
          zakladniPrazec,
          prazceNaDalsiStrune,
          predchoziStrunaPrazec);
      break;
    }
    maxPosun = ZapisPrazceLogic.DEFAULT_POSUN;
    return result;
  }

  /**
   * Povoli vyber s prazdnou strunou. Zakladni prazec neni vzdalen nultemu prazci.
   * @param prazce - seznam prazcu na strune.
   * @param zakladniPrazec - zakladni prazec.
   * @param prazceNaDalsiStrune - seznam prazcu na dalsi strune.
   * @param predchoziStrunaPrazec - prazec predchozi struna.
   * @return Vrati ID prazce.
   */
  private int vyberSPrazdouStrunou(Integer[] prazce,
      int zakladniPrazec,
      List<Integer> prazceNaDalsiStrune,
      int predchoziStrunaPrazec) {
    for (Integer prazec : prazce) {
      if (prazec == ZapisPrazceLogic.PRAZDNA_STRUNA_PRAZEC_ID) {
        return prazec;
      }
    }
    return spolecnaPravidla(prazce,
        zakladniPrazec,
        prazceNaDalsiStrune,
        predchoziStrunaPrazec);
  }

  /**
   * Vyclenuje spolecne chovani pro vsechny pripady. Urcuje priority pro ziskani prazce.
   * @param prazce - seznam prazcu na strune.
   * @param zakladniPrazec - zakladni prazec.
   * @param prazceNaDalsiStrune - seznam prazcu na dalsi strune.
   * @param predchoziStrunaPrazec - prazec predchozi struna.
   * @return Vrati ID prazce.
   */
  private int spolecnaPravidla(Integer[] prazce,
      int zakladniPrazec,
      List<Integer> prazceNaDalsiStrune,
      int predchoziStrunaPrazec) {
    int nalezenyPrazec;
    if ((nalezenyPrazec = ZapisPrazcePriority.isZakladniPrazec(prazce,
        zakladniPrazec)) != ZapisPrazcePriority.NENALEZENY_PRAZEC) {
      return nalezenyPrazec;
    }
    if (zakladniPrazec == ZapisPrazceLogic.DEFAULT_POSUN + 1) {
      if ((nalezenyPrazec = ZapisPrazcePriority.isVzdalenODanouVzdalenost(prazce,
          zakladniPrazec,
          ZapisPrazceLogic.DEFAULT_POSUN - 1)) != ZapisPrazcePriority.NENALEZENY_PRAZEC) {
        return nalezenyPrazec;
      }
    }
    List<Integer> vyhovujiciPrazce = getAllGoodPrazce(zakladniPrazec,
        prazce,
        predchoziStrunaPrazec);
    log.info("vyhovujiciPrazce: {}.",
        vyhovujiciPrazce);
    if (vyhovujiciPrazce.size() == 1) {
      return vyhovujiciPrazce.get(0);
    }
    if (vyhovujiciPrazce.size() == 0) {
      if (maxPosun > ZapisPrazceLogic.DEFAULT_POSUN) {
        if ((nalezenyPrazec = ZapisPrazcePriority.isVzdalenODanouVzdalenost(prazce,
            zakladniPrazec,
            maxPosun)) != ZapisPrazcePriority.NENALEZENY_PRAZEC) {
          return nalezenyPrazec;
        }
      }
    }
    // vyhovujicich prazcu je vice nez 1
    else {
      List<Integer> vyhovujiciPrazceDalsiStruna = getAllGoodPrazce(zakladniPrazec,
          prazceNaDalsiStrune.toArray(new Integer[prazceNaDalsiStrune.size()]),
          predchoziStrunaPrazec);
      log.info("vyhovujiciPrazceDalsiStruna: {}.",
          vyhovujiciPrazceDalsiStruna);
      int result = getPrazecProStrunu(zakladniPrazec,
          vyhovujiciPrazce,
          vyhovujiciPrazceDalsiStruna);
      if (result > ZapisPrazceLogic.NEPOUZITA_STRUNA_PRAZEC_ID) {
        return result;
      }
      else { // hledani prazce do maximalni vzdalenosti
        if ((nalezenyPrazec = ZapisPrazcePriority.isVzdalenODanouVzdalenost(prazce,
            zakladniPrazec,
            ZapisPrazceLogic.DEFAULT_POSUN - 1)) != ZapisPrazcePriority.NENALEZENY_PRAZEC) {
          return nalezenyPrazec;
        }
        if ((nalezenyPrazec = ZapisPrazcePriority.isVzdalenODanouVzdalenost(prazce,
            zakladniPrazec,
            ZapisPrazceLogic.DEFAULT_POSUN)) != ZapisPrazcePriority.NENALEZENY_PRAZEC) {
          return nalezenyPrazec;
        }
      }
    }
    return ZapisPrazceLogic.NEPOUZITA_STRUNA_PRAZEC_ID;
  }

  /**
   * Vyhleda a vraci vsechny vyhovujici prazce na strune.
   * @param zakladniPrazec.
   * @param moznePrazce.
   * @param predchoziStrunaPrazec - prazec na predchozi strune.
   * @return Vraci vsechny vyhovujici prazce na strune.
   */
  private List<Integer> getAllGoodPrazce(int zakladniPrazec,
      Integer[] moznePrazce,
      int predchoziStrunaPrazec) {
    List<Integer> goodPrazce = new ArrayList<Integer>();
    for (Integer prazec : moznePrazce) {
      if ((((predchoziStrunaPrazec - zakladniPrazec) < -1 && (prazec - zakladniPrazec) < 0) ||
          ((predchoziStrunaPrazec - zakladniPrazec) > 1 && (prazec - zakladniPrazec) > 0) ||
          ((predchoziStrunaPrazec - zakladniPrazec) == 1) || ((predchoziStrunaPrazec - zakladniPrazec) == 0) ||
          ((predchoziStrunaPrazec - zakladniPrazec) == -1) || (predchoziStrunaPrazec == zakladniPrazec) || (predchoziStrunaPrazec == ZapisPrazceLogic.NEPOUZITA_STRUNA_PRAZEC_ID)) &&
          (Math.abs(zakladniPrazec - prazec) <= ZapisPrazceLogic.DEFAULT_POSUN || Math.abs(prazec - zakladniPrazec) <= ZapisPrazceLogic.DEFAULT_POSUN)) {
        goodPrazce.add(prazec);
      }
    }
    return goodPrazce;
  }

  /**
   * Porovna dva seznamy moznych prazcu na sousednich strunach a vraci nejvhodnejsi prazec.
   * @param zakladniPrazec - zakladni prazec.
   * @param vyhovujiciPrazce - vyhovujici prazce na prvni strune.
   * @param vyhovujiciPrazceDalsiStruna - vyhovujici prazce na dalsi strune.
   * @return Vraci nejvhodnejsi prazec. Pokud nenajde, vraci -1.
   */
  private int getPrazecProStrunu(int zakladniPrazec,
      List<Integer> vyhovujiciPrazce,
      List<Integer> vyhovujiciPrazceDalsiStruna) {
    int vyslednyPrazec = ZapisPrazceLogic.NEPOUZITA_STRUNA_PRAZEC_ID;
    int maxSoucetVzdalenostiObeStruny = 8;
    int minSoucetVzdalenostiObeStruny = maxSoucetVzdalenostiObeStruny;
    for (Integer prazec : vyhovujiciPrazce) {
      int prazecVetsi = prazec - zakladniPrazec;
      for (Integer prazecDalsiStruna : vyhovujiciPrazceDalsiStruna) {
        int prazecDalsiStrunaVetsi = prazecDalsiStruna - zakladniPrazec;
        if ((prazecVetsi > 0 && prazecDalsiStrunaVetsi >= 0) || (prazecVetsi < 0 && prazecDalsiStrunaVetsi <= 0)) {
          int minSoucet;
          if ((minSoucet = (Math.abs(prazecVetsi) + Math.abs(prazecDalsiStrunaVetsi))) < minSoucetVzdalenostiObeStruny) {
            minSoucetVzdalenostiObeStruny = minSoucet;
            vyslednyPrazec = prazec;
          }
        }
      }
    }
    return vyslednyPrazec;
  }
}